import serial
import serial.tools.list_ports
import platform
import subprocess
import sys
import plotly.graph_objs as go
from IPython.display import display
from .faultier_pb2 import *
import struct
import subprocess
import os

# Get the directory of the current module
MODULE_DIR = os.path.dirname(os.path.realpath(__file__))

def convert_uint8_samples(input: bytes):
    r = []
    for b in input:
        r.append(b/255)
    return r
"""
    This class is used to control the Faultier.

    All functionality that configures the ADC or the glitcher
    require opening the device first, the OpenOCD debug probe
    is always active and does not use the serial interface.
    Because of this the functions using OpenOCD (such as nrf_flash_and_lock)
    are marked as static and can be called without initializing
    the Faultier first.
"""
class Faultier:
    """
    :param path: The path to the serial device. Note that the Faultier exposes
                 two serial devices - the first one is the control channel.
                 
                 On mac this will be /dev/cu.usbmodemfaultier1.
    """

    VID = "2b3e"
    PID = "2343"

    def _open_auto(self):
        self.index = 0
        path = self._find_serial_port(self.index)
        if not path:
            raise Exception("No suitable serial port found.")
        self.device = serial.Serial(path)
        self.device.timeout = 5
        try:
            self._send_hello()
            return
        except self.device.SerialTimeoutException:
            # Attempt second port
            self.index = 1
            path = self._find_serial_port(self.index)
            if not path:
                raise Exception("No suitable serial port found.")
            self.device = serial.Serial(path)
            self.device.timeout = 5
            self._send_hello()

    def _send_hello(self):
        # Send hello command to get protocol version from Faultier
        hello = CommandHello()
        cmd = Command()
        cmd.hello.CopyFrom(hello)
        self._send_protobuf(cmd)
        response = self._check_response()
        if response.hello.version != FAULTIER_VERSION:
            self.device.close()
            raise ValueError(f"Invalid Faultier version: Locally: {FAULTIER_VERSION} - Device: {response.hello.version}")
        

    def __init__(self, path = None):
        """
        """
        if path:
            self.device = serial.Serial(path)
            self.device.timeout = 5
            # Attempt to send hello
            self._send_hello()
        else:
            self._open_auto()

        
        self.default_settings()
    
    def get_serial_path(self):
        """
        The Faultier comes up as two serial ports. The first one is the control channel,
        and the second one is the UART bridge onto the 20-pin connector.
        This function returns the path to the second serial port.
        """
        if(self.index == 0):
            return self._find_serial_port(index=1)
        else:
            return self._find_serial_port(index=0)

    def _find_serial_port(self, index = 0):
        system = platform.system()
        if system == "Windows":
            return self._find_serial_port_windows(index)
        elif system == "Darwin":  # macOS
            return self._find_serial_port_macos(index)
        elif system == "Linux":
            return self._find_serial_port_linux(index)
        else:
            raise Exception(f"Unsupported platform: {system}")
    
    def _find_serial_port_windows(self, index = 0):
        i = 0
        for port in serial.tools.list_ports.comports():
            if self.VID.lower() in port.hwid.lower() and self.PID.lower() in port.hwid.lower():
                if(i == index):
                    return port.device
                i += 1
        return None

    def _find_serial_port_macos(self, index = 0):
        i = 0
        ports = serial.tools.list_ports.comports()
        faultiers = []
        for port in ports:
            if f"USB VID:PID={self.VID.upper()}:{self.PID.upper()}" in port.hwid:
                faultiers.append(port.device)
        if len(faultiers) == 0:
            return None
        
        faultiers.sort()
        return faultiers[index]

    def _find_serial_port_linux(self, index = 0):
        if(index == 0):
            return "/dev/serial/by-id/usb-stacksmashing_Faultier_faultier-if00"
        if(index == 1):
            return "/dev/serial/by-id/usb-stacksmashing_Faultier_faultier-if03"
        return None

    def _read_response(self):
        header = self.device.read(4)
        if(header != b"FLTR"):
            print(header)
            raise ValueError(f"Invalid header received: {header}")
        length_data = self.device.read(4)
        length = struct.unpack("<I", length_data)[0]
        return self.device.read(length)

    def _check_response(self):
        response = self._read_response()
        resp = Response()
        resp.ParseFromString(response)
        if resp.WhichOneof('type') == 'error':
            raise ValueError("Error: " + resp.error.message)
        if resp.WhichOneof('type') == 'trigger_timeout':
            raise ValueError("Trigger timeout!")
        return resp

    def _check_ok(self):
        response = self._read_response()
        resp = Response()
        resp.ParseFromString(response)
        if resp.ok:
            return
        if resp.error:
            raise ValueError("Error: " + resp.error.message)
        else:
            raise ValueError("No OK or Error received.", resp)

    # def captureADC(self):
    #     # TODO will be replaced by command structure
    #     self.device.write(b"A")
    #     response = self._read_response()
    #     resp = ResponseADC()
    #     resp.ParseFromString(response)
    #     return convert_uint8_samples(resp.samples)
    
    def _send_protobuf(self, protobufobj):
        serialized = protobufobj.SerializeToString()
        length = len(serialized)
        # Header
        self.device.write(b"FLTR")
        self.device.write(struct.pack("<I", length))
        self.device.write(serialized)
        self.device.flush()

    def _get_default_settings(self):
        return CommandConfigureGlitcher(
                trigger_type = TRIGGER_NONE,
                trigger_source = TRIGGER_IN_NONE,
                glitch_output = OUT_NONE,
                delay = 0,
                pulse = 0,
                power_cycle_output = OUT_NONE,
                power_cycle_length = 0,
                trigger_pull_configuration = TRIGGER_PULL_NONE)

    def default_settings(self):
        """
        Reset the glitcher configuration to the default settings, namely: No trigger,
        no glitch or power-cycle output, delay and pulse to 0, and the pull-configuration
        of the trigger to none.
        """
        self.glitcher_configuration = self._get_default_settings()

    def configure_adc(self, source, sample_count):
        """
        Configures the ADC of the Faultier. The ADC is filled every-time
        that power_cycle() or glitch() is run and starts running after the
        trigger triggered (or immediately if no trigger is configured).

        :param source: The ADC channel to use

            - `ADC_CROWBAR`: Measure on the Crowbar pin (drain-side)
            - `ADC_MUX0`: Measure on the MUX0 pin
            - `ADX_EXT1`: Measure on the EXT1 pin

        :param sample_count: The number of ADC samples to collect. Maximum is 30000.
        """

        if sample_count > 30000:
            raise ValueError(f"Sample count must be under 30000. Provided {sample_count}.")
        configure_adc = CommandConfigureADC(
            source = source,
            sample_count = sample_count
        )
        cmd = Command()
        cmd.configure_adc.CopyFrom(configure_adc)
        self._send_protobuf(cmd)
        self._check_ok()

    def configure_glitcher(self, trigger_type = None, trigger_source = None, glitch_output = None, delay = None, pulse = None, power_cycle_length= None, power_cycle_output = None, trigger_pull_configuration = None):
        """
        Configures the glitcher, i.e. glitch-output, delay, pulse, etc. It does not Arm
        or cause any other change to IOs until glitch() is called.

        :param delay: The delay between Trigger and Glitch.

        :param pulse: The glitch pulse length.

        :param trigger_type: The type of trigger that should be used

            - `TRIGGER_NONE`: No Trigger. The glitch will wait the delay and then glitch.
            - `TRIGGER_LOW`: Waits for the signal to be low. If the signal is low to begin with this will trigger immediately.
            - `TRIGGER_HIGH`: Waits for the signal to be high. If the signal is low to begin with this will trigger immediately.
            - `TRIGGER_RISING_EDGE`: Waits for a rising edge on the trigger input.
            - `TRIGGER_FALLING_EDGE`: Waits for a falling edge on the trigger input.
            - `TRIGGER_PULSE_POSITIVE`: Wait for a positive pulse on the trigger input.
            - `TRIGGER_PULSE_NEGATIVE`: Wait for a negative pulse on the trigger input.
        
        :param trigger_source: The source - as in physical input - of the trigger.

            - `TRIGGER_IN_NONE`: Ignored, use TRIGGER_NONE to disable triggering.
            - `TRIGGER_IN_EXT0`: Configure EXT0 as digital input and use it for triggering.
            - `TRIGGER_IN_EXT1`: Configure EXT1 as digital input and use it for triggering.

        :param trigger_pulls: The pull up/down configuration of the trigger.

            - `TRIGGER_PULL_NONE`: No pulls applied (default)
            - `TRIGGER_PULL_UP`: Pull-up
            - `TRIGGER_PULL_DOWN`: Pull-down

        :param glitch_output: The glitch-output that will be used.

            - `OUT_CROWBAR`: Route the glitch to the gate of the Crowbar MOSFET.
            - `OUT_MUX0`: Route the glitch to control channel 0/X of the analogue switch (the one exposed on the SMA connector).
            - `OUT_MUX1`: Route the glitch to channel 1/Y of the analogue switch (exposed on 20-pin header).
            - `OUT_MUX2`: Route the glitch to channel 2/Z of the analogue switch (exposed on 20-pin header).
            - `OUT_EXT0`: Route the glitch signal to the EXT0 header. Useful to trigger external tools such as a ChipSHOUTER or laser.
            - `OUT_EXT1`: Route the glitch signal to the EXT1 header. Same as above.
            - `OUT_NONE`: Disable glitch generation. Power-cycle, trigger, ADC & co will still run as regular. Good for trigger testing.

        :param power_cycle: Whether a separate output should be toggled before activating the trigger-delay-glitch pipeline. Useful to restart a target.

            - `OUT_CROWBAR`: Route the power-cycle to the gate of the Crowbar MOSFET.
            - `OUT_MUX0`: Route the power-cycle to control channel 0/X of the analogue switch (the one exposed on the SMA connector).
            - `OUT_MUX1`: Route the power-cycle to channel 1/Y of the analogue switch (exposed on 20-pin header).
            - `OUT_MUX2`: Route the power-cycle to channel 2/Z of the analogue switch (exposed on 20-pin header).
            - `OUT_EXT0`: Route the power-cycle signal to the EXT0 header. Useful to trigger external tools such as a ChipSHOUTER or laser.
            - `OUT_EXT1`: Route the power-cycle signal to the EXT1 header. Same as above.
            - `OUT_NONE`: Disable power-cycle generation.

        :param power_cycle_length: The number of clock-cycles for the power cycle.
        """


        if trigger_type is not None:
            self.glitcher_configuration.trigger_type = trigger_type
        
        if trigger_source is not None:
            self.glitcher_configuration.trigger_source = trigger_source
        
        if glitch_output is not None:
            self.glitcher_configuration.glitch_output = glitch_output
        
        if delay is not None:
            self.glitcher_configuration.delay = delay
        
        if pulse is not None:
            self.glitcher_configuration.pulse = pulse
        
        if power_cycle_length is not None:
            self.glitcher_configuration.power_cycle_length = power_cycle_length
        
        if power_cycle_output is not None:
            self.glitcher_configuration.power_cycle_output = power_cycle_output
        
        if trigger_pull_configuration is not None:
            self.glitcher_configuration.trigger_pull_configuration = trigger_pull_configuration

    def _send_configuration(self, config = None):
        cmd = Command()
        if config:
            cmd.configure_glitcher.CopyFrom(config)
        else:    
            cmd.configure_glitcher.CopyFrom(self.glitcher_configuration)
        self._send_protobuf(cmd)
        self._check_ok()

    def glitch(self, delay = None, pulse = None):
        """
        Perform a glitch. Namely power-cycles (if enabled), then waits for a
        trigger (if configured), then waits for delay-cycles, and then enables
        the configured glitch-output (if any) for pulse-cycles.

        :param delay: Delay between trigger and glitch

        :param pulse: Pulse length for the glitch
        """

        if delay != None:
            self.glitcher_configuration.delay = delay
        if pulse != None:
            self.glitcher_configuration.pulse = pulse
        self._glitch()

    def _glitch(self):
        self._send_configuration()

        cmd = Command()
        cmd.glitch.CopyFrom(CommandGlitch())
        self._send_protobuf(cmd)
        self._check_response()

    def glitch_non_blocking(self, delay = None, pulse = None):
        """
        A non-blocking version of the glitch function. Allows to arm a glitch
        but then still run Python code. Useful for example if your trigger is based
        on another piece of code (i.e. starting a debugger transaction or sending
        something via UART.)

        You MUST call `glitch_check_non_blocking_response` for each glitch_non_blocking
        call, otherwise the communication between the host and the Faultier will desync.
        """

        if delay != None:
            self.glitcher_configuration.delay = delay
        if pulse != None:
            self.glitcher_configuration.pulse = pulse

        self._send_configuration()
        cmd = Command()
        cmd.glitch.CopyFrom(CommandGlitch())
        self._send_protobuf(cmd)
    
    def glitch_check_non_blocking_response(self):
        """
        Reads the response for a non-blocking glitch from the Faultier. This function will
        then cause the error-exceptions such as TriggerTimeout etc.
        """
        self._check_response()

    def swd_check(self):
        """
        Uses the Program header to very quickly check whether an SWD device
        can be found. Useful when for example checking whether a glitch re-enabled
        SWD such as on the STM32 RDP2 to RDP1 glitch.
        """
        cmd = Command()
        cmd.swd_check.CopyFrom(CommandSWDCheck(function = SWD_CHECK_ENABLED))
        self._send_protobuf(cmd)
        response = self._check_response()
        return response.swd_check.enabled

    def nrf52_check(self):
        """
        nRF52 specific check to see whether APPROTECT is disabled/flash can be read.
        """
        cmd = Command()
        cmd.swd_check.CopyFrom(CommandSWDCheck(function = SWD_CHECK_NRF52))
        self._send_protobuf(cmd)
        try:
            response = self._check_response()
        except:
            # TODO check for debug errors only
            return False
            pass
        return response.swd_check.enabled

    def power_cycle(self):
        """
        Power-cycles the target.

        Implementation-wise this actually causes a full glitch to be run, but with
        trigger, delay, pulse, and glitch-output disabled. This means that a power-cycle
        will also fill the ADC.
        """
        config = self._get_default_settings()
        config.power_cycle_output = self.glitcher_configuration.power_cycle_output
        config.power_cycle_length = self.glitcher_configuration.power_cycle_length
        self._send_configuration(config)

        cmd = Command()
        cmd.glitch.CopyFrom(CommandGlitch())
        self._send_protobuf(cmd)
        self._check_response()

    def read_adc(self):
        """
        Receives the current ADC sample-buffer from the device.
        """
        cmd = Command()
        cmd.read_adc.CopyFrom(CommandReadADC())
        self._send_protobuf(cmd)
        response = self._check_response()
        return convert_uint8_samples(response.adc.samples)

    # @staticmethod
    # def nrf_flash_and_lock():
    #     Faultier.nrf_unlock()
    #     print("Programming softdevice...")
        
    #     subprocess.run([
    #         "openocd", "-s", "/usr/local/share/openocd",
    #         "-f", "interface/tamarin.cfg",
    #         "-f", "target/nrf52.cfg",
    #         "-c", f"program {os.path.join(MODULE_DIR, 'example_firmware', 's132_nrf52_7.2.0_softdevice.hex')}; exit"
    #         ])
    #     print("Programming firmware...")
    #     subprocess.run([
    #         "openocd", "-f", "interface/tamarin.cfg", "-f", "target/nrf52.cfg",
    #         "-c", f"program {os.path.join(MODULE_DIR, 'example_firmware', 'nrf52832_xxaa.hex')}; exit"
    #     ])
    #     print("Locking chip...")
    #     Faultier.nrf_lock()

    @staticmethod
    def flash_nrf(path):
        """
        Flashes a connect nRF52 with the provided firmware. Requires OpenOCD to be in path.
        """
        print("Flashing nRF...")
        if not os.path.isfile(path):
            raise Exception(f"File {path} not found.")
        cmd = [
            "openocd",
            "-f", "interface/tamarin.cfg",
            "-f", "target/nrf52.cfg",
            "-c", f"init; nrf52_recover; program {path} verify; reset; exit"
            ]
        try:
            result = subprocess.run(cmd, check=True, text=True, capture_output=True)

            if "Verified OK" in result.stdout or "Verified OK" in result.stderr:
                print("Flashing successful: Verified OK")
            else:
                print("Error: 'Verified OK' not found in OpenOCD output")
                print("Output:", result.stdout)
                print("Errors:", result.stderr)
        except subprocess.CalledProcessError as e:
            print("Error during flashing process:", e)
            print("Output:", e.stdout)
            print("Errors:", e.stderr)


    @staticmethod
    def lock_and_flash_nrf(path):
        """
        Flashes and then APPROTECT-locks a connected nRF52 with the provided firmware.
        Requires OpenOCD to be in path.
        """
        print("Flashing nRF...")
        if not os.path.isfile(path):
            raise Exception(f"File {path} not found.")
        cmd = [
            "openocd",
            "-f", "interface/tamarin.cfg",
            "-f", "target/nrf52.cfg",
            "-c", f"init; nrf52_recover; program {path} verify;  flash fillw 0x10001208 0xFFFFFF00 0x01; reset; exit"
            ]
        try:
            result = subprocess.run(cmd, check=True, text=True, capture_output=True)

            if "Verified OK" in result.stdout or "Verified OK" in result.stderr:
                print("Flashing successful: Verified OK")
            else:
                print("Error: 'Verified OK' not found in OpenOCD output")
                print("Output:", result.stdout)
                print("Errors:", result.stderr)
        except subprocess.CalledProcessError as e:
            print("Error during flashing process:", e)
            print("Output:", e.stdout)
            print("Errors:", e.stderr)


    @staticmethod
    def check_nrf_lock():
        """
        Checks whether the connected device has APPROTECT enabled using OpenOCD.
        Similar to nrf52_check, but much slower and using OpenOCD.
        """
        cmd = [
            "openocd",
            "-f", "interface/tamarin.cfg",
            "-f", "target/nrf52.cfg",
            "-c", f"init; exit"
            ]
        try:
            result = subprocess.run(cmd, check=True, text=True, capture_output=True)
            nrf_msg = "nRF52 device has AP lock engaged"
            if nrf_msg in result.stdout or nrf_msg in result.stderr:
                return True
            else:
                return False
        except subprocess.CalledProcessError as e:
            raise Exception("Failed to check locking status: " + e.stdout + e.stderr)

    @staticmethod
    def lock_nrf():
        """
        Locks the connect nRF52 using APPROTECT. Requires OpenOCD.
        """
        print("Locking nRF...")
        cmd = [
            "openocd",
            "-f", "interface/tamarin.cfg",
            "-f", "target/nrf52.cfg",
            "-c", f"init; flash fillw 0x10001208 0xFFFFFF00 0x01; reset; exit"
            ]
        try:
            result = subprocess.run(cmd, check=True, text=True, capture_output=True)
            print("Chip locked!")
        except subprocess.CalledProcessError as e:
            print("Error during locking process:", e)
            print("Output:", e.stdout)
            print("Errors:", e.stderr)


    @staticmethod
    def unlock_nrf():
        """
        Unlocks the connect nRF52 by running nrf52_recover in OpenOCD.
        """
        print("Unlocking nRF...")
        cmd = [
            "openocd",
            "-f", "interface/tamarin.cfg",
            "-f", "target/nrf52.cfg",
            "-c", f"init; nrf52_recover; exit"
            ]
        try:
            result = subprocess.run(cmd, check=True, text=True, capture_output=True)
            print("Chip unlocked!")
        except subprocess.CalledProcessError as e:
            print("Error during locking process:", e)
            print("Output:", e.stdout)
            print("Errors:", e.stderr)


    @staticmethod
    def nrf_flash():
        Faultier.nrf_unlock()
        print("Programming softdevice...")
        
        subprocess.run([
            "openocd", "-s", "/usr/local/share/openocd",
            "-f", "interface/tamarin.cfg",
            "-f", "target/nrf52.cfg",
            "-c", f"program {os.path.join(MODULE_DIR, 'example_firmware', 's132_nrf52_7.2.0_softdevice.hex')}; exit"
            ])
        print("Programming firmware...")
        subprocess.run([
            "openocd", "-f", "interface/tamarin.cfg", "-f", "target/nrf52.cfg",
            "-c", f"program {os.path.join(MODULE_DIR, 'example_firmware', 'nrf52832_xxaa.hex')}; exit"
        ])

    @staticmethod
    def nrf_lock():
        subprocess.run(["openocd", "-f", "interface/tamarin.cfg", "-f", "target/nrf52.cfg", "-c", "init; reset; halt; flash fillw 0x10001208 0xFFFFFF00 0x01; reset;exit"])

    @staticmethod
    def nrf_unlock():
        print("Unlocking...")
        subprocess.run(["openocd", "-s", "/usr/local/share/openocd", "-f", "interface/tamarin.cfg", "-f", "target/nrf52.cfg", "-c", "init; nrf52_recover; exit"])
        
    def stm32_lock(self):
        import time
        print("This action is not reversible. Locking in 10 seconds, press stop to interrupt.")
        time.sleep(10)



import plotly.graph_objs as go
import serial
import struct
import os
from sys import platform
from IPython.display import clear_output
import random
import time
import matplotlib.pyplot as plt
from tqdm.notebook import trange, tqdm
import pickle

class GlitchData:
    def __init__(self, name, color="gray", alpha=0.3, zorder=1, render=True):
        self.name = name
        self.delays = []
        self.pulses = []
        self.color = color
        self.alpha = alpha
        self.zorder = zorder
        self.render = render
        pass

    def add(self, delay, pulse):
        self.delays.append(delay)
        self.pulses.append(pulse)
    
    def plot_delays(self):
        # Create histogram
        plt.hist(self.delays, bins='auto', color='blue', alpha=0.7, rwidth=0.85)

        # Add title and labels
        plt.title('Distribution of delays')
        plt.xlabel('Delay')
        plt.ylabel('Frequency')

        # Show plot
        plt.grid(axis='y', alpha=0.75)
        plt.show()


    def plot_pulses(self):
        # Create histogram
        plt.hist(self.pulses, bins='auto', color='blue', alpha=0.7, rwidth=0.85)

        # Add title and labels
        plt.title('Distribution of pulses')
        plt.xlabel('Pulse')
        plt.ylabel('Frequency')

        # Show plot
        plt.grid(axis='y', alpha=0.75)
        plt.show()
    

class GlitchDataCollection:
    def __init__(self):
        self.data = {}
        self.min_x = 0
        self.max_x = 0
        self.min_y = 0
        self.max_y = 0
        pass

    def add_data(self, key, name, color="gray", alpha=0.3, zorder=1, render=True):
        self.data[key] = GlitchData(name, color=color, alpha=alpha, zorder=zorder, render=render)
    
    def add(self, key, delay, pulse):
        if not self.min_x or self.min_x > delay:
            self.min_x = delay
        if not self.max_x or self.max_x < delay:
            self.max_x = delay
        if not self.min_y or self.min_y > pulse:
            self.min_y = pulse
        if not self.max_y or self.max_y < pulse:
            self.max_y = pulse
        self.data[key].add(delay, pulse)

    def get_data(self, key):
        return self.data[key]

    def plot(self, x=None, y=None):
        if not x:
            x = [self.min_x, self.max_x]
        if not y:
            y = [self.min_y, self.max_y]
            
        fig, ax = plt.subplots()
        for k in self.data.keys():
            glitch_data = self.data[k]
            alpha = 0.3
            zorder = 1
            ax.scatter(glitch_data.delays, glitch_data.pulses, label=glitch_data.name, alpha=glitch_data.alpha, zorder=glitch_data.zorder, c=glitch_data.color, marker="x")
        ax.legend()
        box = ax.get_position()
        ax.set_position([box.x0, box.y0 + box.height * 0.1,
                        box.width, box.height * 0.9])

        # Put a legend below current axis
        ax.legend(loc='upper center', bbox_to_anchor=(0.5, -0.05),
                fancybox=True, shadow=True, ncol=5)
        ax.grid(True)
        plt.xlim(x[0], x[1])
        plt.ylim(y[0], y[1])
        clear_output(wait=True)
        plt.show()

    def save(self, filename):
        f = open(filename, "wb")
        pickle.dump(self, f)

    @staticmethod
    def load(filename):
        f = open(filename, "rb")
        return pickle.load(f)



import random
from math import gcd

class RandomOrderGenerator:
    def __init__(self, start, end):
        self.n = end - start
        self.start = start
        self.a = self.find_coprime_a()
        self.b = random.randint(0, self.n-1)
        self.current_x = 0

    def find_coprime_a(self):
        # Try to find a coprime number in [n/2, n) and limit the search to 100,000 attempts to ensure fast execution
        attempts = 0
        while attempts < 1000000:
            candidate = random.randint(self.n // 2, self.n - 1)
            if gcd(candidate, self.n) == 1:
                return candidate
            attempts += 1
        raise ValueError("Failed to find a coprime number within 1,000,000 attempts")

    def next_value(self):
        if self.current_x >= self.n:
            raise StopIteration("All values have been visited")
        value = (self.a * self.current_x + self.b) % self.n
        self.current_x += 1
        return self.start + value

    def reset(self):
        self.current_x = 0
        self.b = random.randint(0, self.n-1)



from .Faultier import *
from .FaulterVis import *
from .LivePlot import *
from .RandomOrderGenerator import RandomOrderGenerator
from .GlitchDataCollection import GlitchDataCollection
from .FaultierTool import *



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: faultier.proto
# Protobuf Python Version: 4.25.3
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0e\x66\x61ultier.proto\"\x1f\n\x0f\x43\x61ptureResponse\x12\x0c\n\x04\x64\x61ta\x18\x01 \x01(\x0c\"\x0e\n\x0c\x43ommandHello\"\x10\n\x0e\x43ommandCapture\"\x0f\n\rCommandGlitch\"6\n\x0f\x43ommandSWDCheck\x12#\n\x08\x66unction\x18\x01 \x01(\x0e\x32\x11.SWDCheckFunction\"\xb1\x02\n\x18\x43ommandConfigureGlitcher\x12)\n\x12power_cycle_output\x18\x07 \x01(\x0e\x32\r.GlitchOutput\x12\x1a\n\x12power_cycle_length\x18\x08 \x01(\x05\x12#\n\x0ctrigger_type\x18\x01 \x01(\x0e\x32\r.TriggersType\x12\r\n\x05\x64\x65lay\x18\x02 \x01(\x05\x12\r\n\x05pulse\x18\x03 \x01(\x05\x12&\n\x0etrigger_source\x18\x04 \x01(\x0e\x32\x0e.TriggerSource\x12$\n\rglitch_output\x18\x05 \x01(\x0e\x32\r.GlitchOutput\x12=\n\x1atrigger_pull_configuration\x18\x06 \x01(\x0e\x32\x19.TriggerPullConfiguration\"G\n\x13\x43ommandConfigureADC\x12\x1a\n\x06source\x18\x01 \x01(\x0e\x32\n.ADCSource\x12\x14\n\x0csample_count\x18\x02 \x01(\x05\"\x10\n\x0e\x43ommandReadADC\"\xaa\x02\n\x07\x43ommand\x12\x1e\n\x05hello\x18\x01 \x01(\x0b\x32\r.CommandHelloH\x00\x12\x37\n\x12\x63onfigure_glitcher\x18\x02 \x01(\x0b\x32\x19.CommandConfigureGlitcherH\x00\x12-\n\rconfigure_adc\x18\x03 \x01(\x0b\x32\x14.CommandConfigureADCH\x00\x12\"\n\x07\x63\x61pture\x18\x04 \x01(\x0b\x32\x0f.CommandCaptureH\x00\x12 \n\x06glitch\x18\x05 \x01(\x0b\x32\x0e.CommandGlitchH\x00\x12#\n\x08read_adc\x18\x06 \x01(\x0b\x32\x0f.CommandReadADCH\x00\x12%\n\tswd_check\x18\x07 \x01(\x0b\x32\x10.CommandSWDCheckH\x00\x42\x05\n\x03\x63md\"\x0c\n\nResponseOk\" \n\rResponseError\x12\x0f\n\x07message\x18\x01 \x01(\t\"2\n\rResponseHello\x12!\n\x07version\x18\x01 \x01(\x0e\x32\x10.FaultierVersion\"\x18\n\x16ResponseTriggerTimeout\"\x1e\n\x0bResponseADC\x12\x0f\n\x07samples\x18\x01 \x01(\x0c\"!\n\x0cResponseInfo\x12\x11\n\tfrequency\x18\x01 \x01(\x05\"#\n\x10ResponseSWDCheck\x12\x0f\n\x07\x65nabled\x18\x01 \x01(\x08\"\xe8\x01\n\x08Response\x12\x19\n\x02ok\x18\x01 \x01(\x0b\x32\x0b.ResponseOkH\x00\x12\x1f\n\x05\x65rror\x18\x02 \x01(\x0b\x32\x0e.ResponseErrorH\x00\x12\x1f\n\x05hello\x18\x03 \x01(\x0b\x32\x0e.ResponseHelloH\x00\x12\x1b\n\x03\x61\x64\x63\x18\x04 \x01(\x0b\x32\x0c.ResponseADCH\x00\x12\x32\n\x0ftrigger_timeout\x18\x05 \x01(\x0b\x32\x17.ResponseTriggerTimeoutH\x00\x12&\n\tswd_check\x18\x06 \x01(\x0b\x32\x11.ResponseSWDCheckH\x00\x42\x06\n\x04type*B\n\x0f\x46\x61ultierVersion\x12\x19\n\x15\x46\x41ULTIER_VERSION_ZERO\x10\x00\x12\x14\n\x10\x46\x41ULTIER_VERSION\x10\x01*:\n\x08\x43ommands\x12\r\n\tCMD_RESET\x10\x00\x12\x0e\n\nCMD_GLITCH\x10\x01\x12\x0f\n\x0b\x43MD_CAPTURE\x10\x02*N\n\rTriggerSource\x12\x13\n\x0fTRIGGER_IN_NONE\x10\x00\x12\x13\n\x0fTRIGGER_IN_EXT0\x10\x01\x12\x13\n\x0fTRIGGER_IN_EXT1\x10\x02*s\n\x0cGlitchOutput\x12\x0f\n\x0bOUT_CROWBAR\x10\x00\x12\x0c\n\x08OUT_MUX0\x10\x01\x12\x0c\n\x08OUT_MUX1\x10\x02\x12\x0c\n\x08OUT_MUX2\x10\x03\x12\x0c\n\x08OUT_EXT0\x10\x04\x12\x0c\n\x08OUT_EXT1\x10\x05\x12\x0c\n\x08OUT_NONE\x10\x06*\xae\x01\n\x0cTriggersType\x12\x10\n\x0cTRIGGER_NONE\x10\x00\x12\x10\n\x0cTRIGGER_HIGH\x10\x01\x12\x0f\n\x0bTRIGGER_LOW\x10\x02\x12\x17\n\x13TRIGGER_RISING_EDGE\x10\x03\x12\x18\n\x14TRIGGER_FALLING_EDGE\x10\x04\x12\x1a\n\x16TRIGGER_PULSE_POSITIVE\x10\x05\x12\x1a\n\x16TRIGGER_PULSE_NEGATIVE\x10\x06*8\n\tADCSource\x12\x0f\n\x0b\x41\x44\x43_CROWBAR\x10\x00\x12\x0c\n\x08\x41\x44\x43_MUX0\x10\x01\x12\x0c\n\x08\x41\x44\x43_EXT1\x10\x02*Q\n\x0b\x41uxFunction\x12\x0c\n\x08\x41UX_NONE\x10\x00\x12\x0c\n\x08\x41UX_UART\x10\x01\x12\x13\n\x0f\x41UX_SWD_CHECKER\x10\x02\x12\x11\n\rAUX_SWD_PROBE\x10\x03*>\n\x10SWDCheckFunction\x12\x15\n\x11SWD_CHECK_ENABLED\x10\x00\x12\x13\n\x0fSWD_CHECK_NRF52\x10\x01*]\n\x18TriggerPullConfiguration\x12\x15\n\x11TRIGGER_PULL_NONE\x10\x00\x12\x13\n\x0fTRIGGER_PULL_UP\x10\x01\x12\x15\n\x11TRIGGER_PULL_DOWN\x10\x02\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'faultier_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  _globals['_FAULTIERVERSION']._serialized_start=1323
  _globals['_FAULTIERVERSION']._serialized_end=1389
  _globals['_COMMANDS']._serialized_start=1391
  _globals['_COMMANDS']._serialized_end=1449
  _globals['_TRIGGERSOURCE']._serialized_start=1451
  _globals['_TRIGGERSOURCE']._serialized_end=1529
  _globals['_GLITCHOUTPUT']._serialized_start=1531
  _globals['_GLITCHOUTPUT']._serialized_end=1646
  _globals['_TRIGGERSTYPE']._serialized_start=1649
  _globals['_TRIGGERSTYPE']._serialized_end=1823
  _globals['_ADCSOURCE']._serialized_start=1825
  _globals['_ADCSOURCE']._serialized_end=1881
  _globals['_AUXFUNCTION']._serialized_start=1883
  _globals['_AUXFUNCTION']._serialized_end=1964
  _globals['_SWDCHECKFUNCTION']._serialized_start=1966
  _globals['_SWDCHECKFUNCTION']._serialized_end=2028
  _globals['_TRIGGERPULLCONFIGURATION']._serialized_start=2030
  _globals['_TRIGGERPULLCONFIGURATION']._serialized_end=2123
  _globals['_CAPTURERESPONSE']._serialized_start=18
  _globals['_CAPTURERESPONSE']._serialized_end=49
  _globals['_COMMANDHELLO']._serialized_start=51
  _globals['_COMMANDHELLO']._serialized_end=65
  _globals['_COMMANDCAPTURE']._serialized_start=67
  _globals['_COMMANDCAPTURE']._serialized_end=83
  _globals['_COMMANDGLITCH']._serialized_start=85
  _globals['_COMMANDGLITCH']._serialized_end=100
  _globals['_COMMANDSWDCHECK']._serialized_start=102
  _globals['_COMMANDSWDCHECK']._serialized_end=156
  _globals['_COMMANDCONFIGUREGLITCHER']._serialized_start=159
  _globals['_COMMANDCONFIGUREGLITCHER']._serialized_end=464
  _globals['_COMMANDCONFIGUREADC']._serialized_start=466
  _globals['_COMMANDCONFIGUREADC']._serialized_end=537
  _globals['_COMMANDREADADC']._serialized_start=539
  _globals['_COMMANDREADADC']._serialized_end=555
  _globals['_COMMAND']._serialized_start=558
  _globals['_COMMAND']._serialized_end=856
  _globals['_RESPONSEOK']._serialized_start=858
  _globals['_RESPONSEOK']._serialized_end=870
  _globals['_RESPONSEERROR']._serialized_start=872
  _globals['_RESPONSEERROR']._serialized_end=904
  _globals['_RESPONSEHELLO']._serialized_start=906
  _globals['_RESPONSEHELLO']._serialized_end=956
  _globals['_RESPONSETRIGGERTIMEOUT']._serialized_start=958
  _globals['_RESPONSETRIGGERTIMEOUT']._serialized_end=982
  _globals['_RESPONSEADC']._serialized_start=984
  _globals['_RESPONSEADC']._serialized_end=1014
  _globals['_RESPONSEINFO']._serialized_start=1016
  _globals['_RESPONSEINFO']._serialized_end=1049
  _globals['_RESPONSESWDCHECK']._serialized_start=1051
  _globals['_RESPONSESWDCHECK']._serialized_end=1086
  _globals['_RESPONSE']._serialized_start=1089
  _globals['_RESPONSE']._serialized_end=1321
# @@protoc_insertion_point(module_scope)



from IPython.display import HTML
import xml.etree.ElementTree as ET
import tempfile
import os
import urllib

MODULE_DIR = os.path.dirname(os.path.realpath(__file__))

def update_text_fill(text_element, new_fill_color):
    """
    Updates the "fill" property of an SVG text element's style attribute.

    Args:
    text_element (ET.Element): The SVG text element to modify.
    new_fill_color (str): The new color to set for the fill property.
    """
    # Check if the element has a 'style' attribute
    if 'style' in text_element.attrib:
        # Split the style attribute into individual properties
        styles = text_element.get('style').split(';')
        new_styles = []
        fill_updated = False
        for style in styles:
            if style.strip().startswith('fill:'):
                # Update the fill color
                new_styles.append(f'fill:{new_fill_color}')
                fill_updated = True
            else:
                new_styles.append(style)
        if not fill_updated:
            # If fill was not previously set, add it
            new_styles.append(f'fill:{new_fill_color}')
        # Update the style attribute
        text_element.set('style', ';'.join(new_styles))
    else:
        # If there's no style attribute, simply add one with the fill color
        text_element.set('style', f'fill:{new_fill_color}')


class FaultierVis:
    def __init__(self, svg_file_path):
        ET.register_namespace('', 'http://www.w3.org/2000/svg')
         # Parse the SVG file
        tree = ET.parse(svg_file_path)
        self.root = tree.getroot()

        # SVG files have namespaces, so we need to handle them.
        self.namespaces = {'svg': 'http://www.w3.org/2000/svg'}

    def change_fill(self, group_id, fill):
        g_element = self.root.find(f".//*[@id='{group_id}']", self.namespaces)

        if g_element is not None:
            # Find the 'text' element within the 'g' element. Assuming it's the first 'text' element for simplicity.
            text_element = g_element.find('.//svg:text', self.namespaces)
            
            if text_element is not None:
                update_text_fill(text_element, fill)
            else:
                print(f"No 'text' element found within the group '{group_id}'.")
                return None
        else:
            print(f"No element with ID '{group_id}' found in the SVG file.")
            return None

    def replace_text(self, group_id, text, fill=None):
        # Find the 'g' element by ID
        g_element = self.root.find(f".//*[@id='{group_id}']", self.namespaces)

        if g_element is not None:
            # Find the 'text' element within the 'g' element. Assuming it's the first 'text' element for simplicity.
            text_element = g_element.find('.//svg:text', self.namespaces)
            
            if text_element is not None:
                # Update the entire text content directly if you want to replace everything inside <text>
                text_element.text = text

                # Optionally, clear existing tspan elements (if you want to remove them)
                for tspan in text_element.findall('.//svg:tspan', self.namespaces):
                    text_element.remove(tspan)

                if fill:
                    update_text_fill(text_element, fill)

                # Convert the updated SVG tree back to a string
                svg_string = ET.tostring(self.root, encoding='unicode')

                # Optionally, strip out the xml declaration if present
                # svg_string = svg_string.replace('<?xml version=\'1.0\' encoding=\'unicode\'?>', '')

                return svg_string
            else:
                print(f"No 'text' element found within the group '{group_id}'.")
                return None
        else:
            print(f"No element with ID '{group_id}' found in the SVG file.")
            return None   


    def create_tag(self):
        # fd, path = tempfile.mkstemp(suffix=".svg", text=True)
        # f = os.fdopen(fd, 'w')
        # print(f)
        svg_string = ET.tostring(self.root, encoding='unicode')

        # URL-encode the SVG string
        encoded_svg = urllib.parse.quote(svg_string)

        # Create the <img> tag with the encoded SVG as the src
        img_tag = f'<img width=800 src="data:image/svg+xml,{encoded_svg}" alt="Inline SVG" />'

        return img_tag

    def show(self):
        display(HTML(self.create_tag()))

    @staticmethod
    def show_stm32_glitch_configuration():
        fv = FaultierVis(MODULE_DIR + "/docs/topview.svg")
        fv.replace_text("text_crowbar", "To STM32 VCore")
        fv.replace_text("text_mux0", "To STM32 VCC")
        fv.replace_text("text_ext0", "To STM32 RST")
        fv.replace_text("text_ext1", "Unused", fill="#999")
        fv.show()
        fv = FaultierVis(MODULE_DIR + "/docs/sideview.svg")
        fv.change_fill("text_swd_vcc", "#555")
        for i in range(0, 10):
            fv.change_fill(f"text_io{i}", "#555")
        for i in range(1, 3):
            fv.change_fill(f"text_m{i}_out", "#555")
            fv.change_fill(f"text_m{i}_in0", "#555")
            fv.change_fill(f"text_m{i}_in1", "#555")
        fv.change_fill(f"text_5v", "#555")
        # fv.replace_text(f"text_3.3v", "3.3V", "#555")
        fv.change_fill(f"text_gnd_top", "#555")
        fv.change_fill(f"text_gnd_bottom", "#555")
        # fv.replace_text("text_mux0", "To STM32 VCC")
        # fv.replace_text("text_ext0", "Unused", fill="#999")
        # fv.replace_text("text_ext1", "Unused", fill="#999")
        fv.show()
    
    @staticmethod
    def show_nrf52_glitch_configuration():
        fv = FaultierVis(MODULE_DIR + "/docs/topview.svg")
        fv.replace_text("text_crowbar", "To nRF52 VCore")
        fv.replace_text("text_mux0", "To nRF52 VCC")
        fv.replace_text("text_ext0", "Unused", fill="#999")
        fv.replace_text("text_ext1", "Unused", fill="#999")
        fv.show()
        fv = FaultierVis(MODULE_DIR + "/docs/sideview.svg")
        fv.change_fill("text_swd_vcc", "#555")
        for i in range(0, 10):
            fv.change_fill(f"text_io{i}", "#555")
        for i in range(1, 3):
            fv.change_fill(f"text_m{i}_out", "#555")
            fv.change_fill(f"text_m{i}_in0", "#555")
            fv.change_fill(f"text_m{i}_in1", "#555")
        fv.change_fill(f"text_5v", "#555")
        # fv.replace_text(f"text_3.3v", "3.3V", "#555")
        fv.change_fill(f"text_gnd_top", "#555")
        fv.change_fill(f"text_gnd_bottom", "#555")
        # fv.replace_text("text_mux0", "To STM32 VCC")
        # fv.replace_text("text_ext0", "Unused", fill="#999")
        # fv.replace_text("text_ext1", "Unused", fill="#999")
        fv.show()



#!/usr/bin/env python
import sys
import os
try:
    import faultier
except:
    print("Can't import faultier - make sure you install the latest faultier version with pip3 install -U faultier")
    sys.exit(1)
import argparse
import subprocess

# Define your functions, all accepting 'args' even if they don't use it
def faultier_nrf52_test(args=None):
    print("Running NRF test...")

    
def faultier_nrf52_flash(file_path):
    """
    Flashes a connect nRF52 with the provided firmware. Requires OpenOCD to be in path.
    """
    print("Flashing nRF...")
    if not os.path.isfile(file_path):
        raise Exception(f"File {file_path} not found.")
    cmd = [
        "openocd",
        "-f", "interface/tamarin.cfg",
        "-f", "target/nrf52.cfg",
        "-c", f"init; nrf52_recover; program {file_path} verify; reset; exit"
        ]
    try:
        result = subprocess.run(cmd, check=True, text=True, capture_output=True)

        if "Verified OK" in result.stdout or "Verified OK" in result.stderr:
            print("Flashing successful: Verified OK")
        else:
            print("Error: 'Verified OK' not found in OpenOCD output")
            print("Output:", result.stdout)
            print("Errors:", result.stderr)
    except subprocess.CalledProcessError as e:
        print("Error during flashing process:", e)
        print("Output:", e.stdout)
        print("Errors:", e.stderr)

def openocd_program(config, path):
    if not os.path.isfile(path):
        raise Exception(f"File {path} not found.")
    if " " in path:
        raise Exception(f"Path contains spaces - unsupported.")
    if ";" in path:
        raise Exception(f"Path contains semicolon - unsupported.")
    cmd = [
        "openocd",
        "-f", "interface/tamarin.cfg",
        "-f", f"target/{config}.cfg",
        "-c", f"init; program {path} verify; reset; exit"
    ]
    try:
        result = subprocess.run(cmd, check=True, text=True, capture_output=True)

        if "Verified OK" in result.stdout or "Verified OK" in result.stderr:
            print("Flashing successful: Verified OK")
        else:
            print("Error: 'Verified OK' not found in OpenOCD output")
            print("Output:", result.stdout)
            print("Errors:", result.stderr)
    except subprocess.CalledProcessError as e:
        print("Error during flashing process:", e)
        print("Output:", e.stdout)
        print("Errors:", e.stderr)


    pass

def faultier_nrf52_lock(args=None):
    print("Locking nRF...")
    cmd = [
    "openocd",
    "-f", "interface/tamarin.cfg",
    "-f", "target/nrf52.cfg",
    "-c", f"init; reset halt; flash fillw 0x10001208 0xFFFFFF00 0x01; reset; exit"
    ]
    try:
        result = subprocess.run(cmd, check=True, text=True, capture_output=True)
        print("Chip locked!")
    except subprocess.CalledProcessError as e:
        print("Error during locking process:", e)
        print("Output:", e.stdout)
        print("Errors:", e.stderr)

    # faultier.Faultier.lock_nrf()

def faultier_nrf52_unlock(args=None):
    print("NRF unlocked.")
    faultier.Faultier.unlock_nrf()

def faultier_nrf52_flash(file_path):
    print(f"Flashing nRF52 with {file_path}")
    openocd_program("nrf52", file_path)

def faultier_stm32_test(args=None):
    print("Running STM32 test...")

def faultier_stm32_rdp1(args=None):
    print("STM32 RDP1 enabled.")

def faultier_stm32_rdp2(really):
    if really:
        print("STM32 RDP2 enabled.")
    else:
        print("You must confirm RDP2 with --really flag.")

def faultier_stm32_rdp0(args=None):
    print("STM32 RDP0 enabled.")

def faultier_stm32_flash(file_path):
    print(f"Flashing STM32 with {file_path}")
    openocd_program("stm32f4x", file_path)

def faultier_test(args=None):
    print("Running test...")

def main():
    # Set up argparse
    parser = argparse.ArgumentParser(description="Faultier tool")

    # Create subparsers for the different modes
    subparsers = parser.add_subparsers(dest="mode", help="")

    # NRF mode
    nrf_parser = subparsers.add_parser("nrf52", help="nRF52 commands")
    nrf_subparsers = nrf_parser.add_subparsers(dest="command", help="NRF commands")

    nrf_subparsers.add_parser("lock", help="Lock NRF").set_defaults(func=faultier_nrf52_lock)
    nrf_subparsers.add_parser("unlock", help="Unlock NRF").set_defaults(func=faultier_nrf52_unlock)

    flash_nrf_parser = nrf_subparsers.add_parser("flash", help="Flash NRF with a file")
    flash_nrf_parser.add_argument("file", help="File path for flashing NRF")
    flash_nrf_parser.set_defaults(func=lambda args: faultier_nrf52_flash(args.file))

    # STM32 mode
    stm32_parser = subparsers.add_parser("stm32", help="STM32 commands")
    stm32_subparsers = stm32_parser.add_subparsers(dest="command", help="STM32 commands")

    stm32_subparsers.add_parser("rdp1", help="Enable RDP1").set_defaults(func=faultier_stm32_rdp1)
    rdp2_parser = stm32_subparsers.add_parser("rdp2", help="Enable RDP2")
    rdp2_parser.add_argument("--really", action="store_true", help="Confirm enabling RDP2")
    rdp2_parser.set_defaults(func=lambda args: faultier_stm32_rdp2(args.really))

    stm32_subparsers.add_parser("rdp0", help="Enable RDP0").set_defaults(func=faultier_stm32_rdp0)

    flash_stm32_parser = stm32_subparsers.add_parser("flash", help="Flash STM32 with a file")
    flash_stm32_parser.add_argument("file", help="File path for flashing STM32")
    flash_stm32_parser.set_defaults(func=lambda args: faultier_stm32_flash(args.file))

    # Parse arguments and check for subcommand
    args = parser.parse_args()

    # Show help if no subcommand was provided
    if args.mode == "nrf52" and args.command is None:
        nrf_parser.print_help()
        sys.exit(1)
    elif args.mode == "stm32" and args.command is None:
        stm32_parser.print_help()
        sys.exit(1)

    # Call the function assigned to the subcommand
    if hasattr(args, 'func'):
        args.func(args)
    else:
        parser.print_help()


if __name__ == "__main__":
    main()


import time        
import plotly.graph_objs as go
def update_vline_position(fig, old_x, new_x):
    for shape in fig.layout.shapes:
        if shape.type == 'line' and shape.x0 == old_x and shape.x1 == old_x:
            shape.update(x0=new_x, x1=new_x)
            break

class LivePlot:
    def __init__(self):
        self.fig = go.FigureWidget(data=[go.Scatter(y=[])])
        self.fig.update_layout(yaxis=dict(range=[0, 1]))
        self.fig.update_layout(
            margin=dict(l=20, r=20, t=20, b=20),
        )
        # vline = self.fig.add_vline(x=200, line_width=1, line_dash="dash", line_color="red")
        self.vline_x = 200
        display(self.fig)
    
    def update(self, data):
        """
        Updates the live figure.

        :param data: Takes the data in the format [5, 3, 2, 1, ...]
        """

        self.fig.data[0].y = data

    def update_vline(self, x):
        """
        Updates the vertical marker on the graph.

        :param data: Takes in the X position on which to place the marker.
        """

        update_vline_position(self.fig, old_x=self.vline_x, new_x=x)
        self.vline_x = x


class LiveMarkerPlotOld:
    def __init__(self):
        self.fig = go.FigureWidget(data=[
            go.Scatter(x=[0], y=[0], mode="markers"),
            go.Scatter(x=[0], y=[0], mode="markers")
        ])
        # self.fig.update_layout(yaxis=dict(range=[0, 1]))
        self.fig.update_layout(
            margin=dict(l=20, r=20, t=20, b=20),
            xaxis_title="Delay",
            yaxis_title="Pulse"
        )
        # vline = self.fig.add_vline(x=200, line_width=1, line_dash="dash", line_color="red")
        self.vline_x = 200
        display(self.fig)
    
    def update(self, data):
        """
        Updates the live figure.

        :param data: Takes the scatters in the format [[4000, 8], [3950, 4]]
        """
        
        x_values = [point[0] for point in data[0]]
        y_values = [point[1] for point in data[0]]

        self.fig.data[0].x = x_values
        self.fig.data[0].y = y_values

        x_values = [point[0] for point in data[1]]
        y_values = [point[1] for point in data[1]]

        self.fig.data[1].x = x_values
        self.fig.data[1].y = y_values


class LiveMarkerPlot:
    def __init__(self, gdc, x_range=None, y_range=None):
        """
        Initializes the LiveMarkerPlot.

        :param gdc: The data collection object.
        :param x_range: The range for the x-axis as a tuple (min, max), or None to auto-scale.
        :param y_range: The range for the y-axis as a tuple (min, max), or None to auto-scale.
        """
        self.gdc = gdc

        figure_widget_data = []
        for key in gdc.data:
            if gdc.data[key].render:
                figure_widget_data.append(
                    go.Scatter(x=[], y=[], mode="markers", name=gdc.data[key].name, marker_color=gdc.data[key].color),
                )

        self.fig = go.FigureWidget(data=figure_widget_data)

        # Set up x and y axis ranges if provided
        layout_options = {
            "margin": dict(l=20, r=20, t=20, b=20),
            "xaxis_title": "Delay",
            "yaxis_title": "Pulse"
        }
        if x_range:
            layout_options["xaxis"] = dict(range=x_range)
        if y_range:
            layout_options["yaxis"] = dict(range=y_range)

        self.fig.update_layout(**layout_options)
        self.vline_x = 200
        self.last_update = time.time()
        display(self.fig)


    
    def slow_update(self):
        if((time.time() - self.last_update) < 3):
            return
        self.update()

    def update(self):
        """
        Updates the live figure.

        :param data: Takes the scatters in the format [[4000, 8], [3950, 4]]
        """
        self.last_update = time.time()
        i = 0
        for k in self.gdc.data:
            data = self.gdc.get_data(k)
            if not data.render:
                continue
            x_values = data.delays
            y_values = data.pulses
            self.fig.data[i].x = x_values
            self.fig.data[i].y = y_values
            i += 1
